"""Send To Kindle."""

import base64
import dataclasses
import hashlib
import json
import os
import urllib.parse
import urllib.request
from pathlib import Path
from typing import Any, BinaryIO, List, Mapping, TextIO, Union

from stkclient import api, model, signer

OwnedDevice = model.OwnedDevice


@dataclasses.dataclass()
class Client:
    """Supports listing devices and sending files to specific devices."""

    _device_info: model.DeviceInfo
    _signer: signer.Signer = dataclasses.field(init=False, repr=False)

    def __post_init__(self) -> None:
        """Initialize _signer."""
        self._signer = signer.Signer.from_device_info(self._device_info)

    def get_owned_devices(self) -> List[OwnedDevice]:
        """Returns a list of kindle devices owned by the end-user.

        Returns:
            List of OwnedDevice instances.
        """
        return api.get_list_of_owned_devices(self._signer).owned_devices

    def send_file(
        self,
        file_path: Path,
        target_device_serial_numbers: List[str],
        *,
        author: str,
        title: str,
        format: str,
    ) -> str:
        """Sends a file to the specified kindle devices.

        Args:
            file_path: The file to send
            target_device_serial_numbers: The devices to receive the file.
            author: The author of the document.
            title: The title of the document.
            format: The format of the document.

        Returns:
            sku identifier assigned by amazon.
        """
        file_size = file_path.stat().st_size
        upload = api.get_upload_url(self._signer, file_size)
        with open(file_path, "rb") as f:
            api.upload_file(upload.upload_url, file_size, f)
        ret = api.send_to_kindle(
            self._signer,
            upload.stk_token,
            target_device_serial_numbers,
            author=author,
            title=title,
            format=format,
        )
        return ret.sku

    def logout(self) -> None:
        """Logs out the client."""
        api.logout(self._signer)

    @staticmethod
    def load(fp: Union[TextIO, BinaryIO]) -> "Client":
        """Deserializes a client from a file-like object."""
        return Client._from_dict(json.load(fp))

    @staticmethod
    def loads(s: str) -> "Client":
        """Deserializes a client from a string."""
        return Client._from_dict(json.loads(s))

    @staticmethod
    def _from_dict(s: Mapping[str, Any]) -> "Client":
        if s.get("version") != 1:
            raise ValueError("Invalid version")
        return Client(model.DeviceInfo.from_dict(s.get("device_info", {})))

    def dump(self, fp: TextIO) -> None:
        """Serializes the client into a file-like object."""
        return json.dump(self._to_dict(), fp)

    def dumps(self) -> str:
        """Serializes the client into a string."""
        return json.dumps(self._to_dict())

    def _to_dict(self) -> Mapping[str, Any]:
        return {"version": 1, "device_info": dataclasses.asdict(self._device_info)}


class OAuth2:
    """Authenticates an end-user using amazon's OAuth2."""

    def __init__(self) -> None:
        """Constructs an OAuth2."""
        self._verifier = _base64_url_encode(os.urandom(32))

    def get_signin_url(self) -> str:
        """Gets the signin URL. Open in a web browser to start authentication."""
        challenge = _base64_url_encode(_sha256(self._verifier.encode("utf-8")))
        q = {
            "openid.claimed_id": "http://specs.openid.net/auth/2.0/identifier_select",
            "openid.ns.oa2": "http://www.amazon.com/ap/ext/oauth/2",
            "openid.ns": "http://specs.openid.net/auth/2.0",
            "openid.identity": "http://specs.openid.net/auth/2.0/identifier_select",
            "openid.oa2.client_id": "device:658490dfb190e494030082836775981fa23be0c2425441860352ba0f55915b43002d",
            "openid.mode": "checkid_setup",
            "openid.oa2.scope": "device_auth_access",
            "openid.oa2.response_type": "code",
            "openid.oa2.code_challenge": challenge,
            "openid.oa2.code_challenge_method": "S256",
            "openid.return_to": "https://www.amazon.com/gp/sendtokindle",
            "openid.ns.pape": "http://specs.openid.net/extensions/pape/1.0",
            "openid.pape.max_auth_age": "0",
            "accountStatusPolicy": "P1",
            "openid.assoc_handle": "amzn_device_na",
            "pageId": "amzn_device_common_dark",
            "disableLoginPrepopulate": "1",
        }
        return "https://www.amazon.com/ap/signin?" + urllib.parse.urlencode(q)

    def create_client(self, redirect_url: str) -> Client:
        """Creates a client with the authorization code from the redirect URL.

        Args:
            redirect_url: The final oauth redirect URL.

        Returns:
            Client instance.
        """
        code = _parse_authorization_code(redirect_url)
        access_token = api.token_exchange(code, self._verifier)
        device_info = api.register_device_with_token(access_token)
        return Client(device_info)


def _parse_authorization_code(redirect_url: str) -> str:
    """Parse authorization_code from an OAuth2 redirect URL.

    Example:
        >>> from stkclient import _parse_authorization_code
        >>> _parse_authorization_code('https://www.amazon.com/gp/sendtokindle?openid.assoc_handle=amzn_device_na&openid.oa2.authorization_code=ANTMNkABICRQwomAuQbjuZMn&')
        'ANTMNkABICRQwomAuQbjuZMn'
    """
    u = urllib.parse.urlparse(redirect_url)
    q = urllib.parse.parse_qs(u.query)
    return q["openid.oa2.authorization_code"][0]


def _base64_url_encode(s: bytes) -> str:
    """Base64 encode.

    Example:
        >>> from stkclient import _base64_url_encode
        >>> _base64_url_encode(b'foo')
        'Zm9v'
    """
    return base64.b64encode(s, b"-_").rstrip(b"=").decode("utf8")


def _sha256(s: bytes) -> bytes:
    m = hashlib.sha256()
    m.update(s)
    return m.digest()


__all__ = ["OAuth2", "OwnedDevice", "Client"]



"""Implements RSA request signing for amazon APIs."""

import base64
import datetime
import hashlib
from dataclasses import dataclass
from typing import Optional, cast

import rsa
from rsa import core, transform

from .model import DeviceInfo


@dataclass(frozen=True, repr=False)
class Signer:
    """Implements RSA request signing for amazon APIs.

    Attributes:
        device_private_key: The private key used to generate the X-ADP-Request-Digest header.
        adp_token: The value to be included in the X-ADP-Authentication-Token header.
    """

    device_private_key: rsa.PrivateKey
    adp_token: str

    @staticmethod
    def from_device_info(d: DeviceInfo) -> "Signer":
        """Constructs a Signer instance from a DeviceInfo object."""
        k = rsa.PrivateKey.load_pkcs1(d.device_private_key.encode("utf-8"))
        return Signer(
            device_private_key=cast(rsa.PrivateKey, k),
            adp_token=d.adp_token,
        )

    def digest_header_for_request(
        self, method: str, path: str, post_data: str, signing_date: Optional[str] = None
    ) -> str:
        """Computes the request digest to be included in the X-ADP-Request-Digest header.

        Args:
            method: The HTTP method of the request
            path: The path portion of the URL being requested (starting with /).
            post_data: The HTTP request body if present, or an empty string
            signing_date: The date included in the signature. If null, the current date is used.

        Returns:
            The request digest to be included in the X-ADP-Request-Digest header.
        """
        if signing_date is None:
            signing_date = _get_signing_date()
        sig_data = self._make_digest_data_for_request(method, path, post_data, signing_date)
        digest = _sha256(sig_data)
        pad = bytes.fromhex(
            "01ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff00"
        )
        padded = pad + digest
        payload = transform.bytes2int(padded)
        encrypted = core.encrypt_int(payload, self.device_private_key.d, self.device_private_key.n)
        encrypted_bytes = transform.int2bytes(encrypted, 256)
        bytes64 = base64.b64encode(encrypted_bytes).decode("utf-8")
        return f"{bytes64}:{signing_date}"

    def _make_digest_data_for_request(
        self, method: str, path: str, post_data: str, signing_date: Optional[str] = None
    ) -> bytes:
        if signing_date is None:
            signing_date = _get_signing_date()
        sig_data = "\n".join([method, path, signing_date, post_data, self.adp_token])
        return sig_data.encode("utf-8")


# 2021-10-09T05:02:38Z
def _get_signing_date() -> str:
    return (
        datetime.datetime.utcnow()
        .replace(microsecond=0, tzinfo=datetime.timezone.utc)
        .isoformat()
        .replace("+00:00", "Z")
    )


def _sha256(s: bytes) -> bytes:
    m = hashlib.sha256()
    m.update(s)
    return m.digest()



"""Send to Kindle API response and domain objects."""

from dataclasses import dataclass, field, fields
from typing import Any, List, Mapping, Optional

try:
    from defusedxml.ElementTree import fromstring as xml_parse
except ImportError:
    from xml.etree.ElementTree import fromstring as xml_parse  # noqa: S405


@dataclass(frozen=True)
class DeviceInfo:
    """Contains authentication information for the client as well as user metadata.

    Attributes:
        device_private_key: The private key used to generate the X-ADP-Request-Digest header.
        adp_token: The value to be included in the X-ADP-Authentication-Token header.
        device_type: An opaque identifier.
        given_name: The end-user's given name.
        name: The end-user's full name.
        account_pool: "Amazon"
        user_directed_id: An opaque identifier.
        user_device_name: The end-user's name for their device.
        home_region: "NA"
    """

    device_private_key: str = field(repr=False)
    adp_token: str = field(repr=False)
    device_type: str
    given_name: str
    name: str
    account_pool: str
    user_directed_id: str
    user_device_name: str
    home_region: Optional[str] = None

    @staticmethod
    def from_dict(d: Mapping[str, Any]) -> "DeviceInfo":
        """Constructs a DeviceInfo from a dictionary of attributes, ignoring unknown fields."""
        fieldnames = {f.name for f in fields(DeviceInfo)}
        return DeviceInfo(**{k: v for k, v in d.items() if k in fieldnames})

    @staticmethod
    def from_xml(xml: bytes) -> "DeviceInfo":
        """Constructs a DeviceInfo from an XML string."""
        res = xml_parse(xml)  # noqa S314
        info = {}
        for el in res:
            info[el.tag] = el.text
        return DeviceInfo.from_dict(info)


@dataclass(frozen=True)
class OwnedDevice:
    """Represents a user device supported by send-to-kindle.

    Attributes:
        device_capabilities: Mapping of capability name to boolean
        device_name: Human readable device name specified by the end user.
        device_serial_number: Unique ID of this device.
    """

    device_capabilities: Mapping[str, bool]
    device_name: str
    device_serial_number: str

    @staticmethod
    def from_dict(d: Mapping[str, Any]) -> "OwnedDevice":
        """Constructs an OwnedDevice from a dictionary of attributes, ignoring unknown fields."""
        return OwnedDevice(
            device_capabilities=d["deviceCapabilities"],
            device_name=d["deviceName"],
            device_serial_number=d["deviceSerialNumber"],
        )


@dataclass(frozen=True)
class GetOwnedDevicesResponse:
    """Response from get_list_of_owned_devices.

    Attributes:
        owned_devices: List of owned devices.
        status_code: 0.
    """

    owned_devices: List[OwnedDevice]
    status_code: int

    @staticmethod
    def from_dict(d: Mapping[str, Any]) -> "GetOwnedDevicesResponse":
        """Constructs a GetOwnedDevicesResponse from a dictionary of attributes."""
        return GetOwnedDevicesResponse(
            owned_devices=[OwnedDevice.from_dict(v) for v in d["ownedDevices"]],
            status_code=d["statusCode"],
        )


@dataclass(frozen=True)
class GetUploadUrlResponse:
    """Response from get_upload_url.

    Attributes:
        expiry_time: When this URL expires.
        status_code: 0
        stk_token: Unique identifier for this url.
        upload_url: The upload URL.
    """

    expiry_time: int
    status_code: int
    stk_token: str
    upload_url: str

    @staticmethod
    def from_dict(d: Mapping[str, Any]) -> "GetUploadUrlResponse":
        """Constructs a GetUploadUrlResponse from a dictionary of attributes."""
        return GetUploadUrlResponse(
            expiry_time=d["expiryTime"],
            status_code=d["statusCode"],
            stk_token=d["stkToken"],
            upload_url=d["uploadUrl"],
        )


@dataclass(frozen=True)
class SendToKindleResponse:
    """Response from send_to_kindle.

    Attributes:
        sku: Opaque ID
        status_code: 0
    """

    sku: str
    status_code: int

    @staticmethod
    def from_dict(d: Mapping[str, Any]) -> "SendToKindleResponse":
        """Constructs a SendToKindleResponse from a dictionary of attributes."""
        return SendToKindleResponse(
            sku=d["sku"],
            status_code=d["statusCode"],
        )



"""Typed wrapper functions for the amazon auth and stk APIs."""

import http.client
import json
import urllib.error
import urllib.parse
import urllib.request
from typing import IO, Any, List, Mapping, Optional

from stkclient.model import (
    DeviceInfo,
    GetOwnedDevicesResponse,
    GetUploadUrlResponse,
    SendToKindleResponse,
)
from stkclient.signer import Signer

DEFAULT_CLIENT_INFO = {
    "appName": "ShellExtension",
    "appVersion": "1.1.1.253",
    "os": "MacOSX_10.14.6_x64",
    "osArchitecture": "x64",
}


class APIError(ValueError):
    """Represents errors returned in HTTP response of the API."""

    def __init__(self, msg: str, body: Optional[bytes]):
        """Construct an APIError with a given message and response body."""
        if body is not None:
            try:
                body = json.loads(body)
            except json.JSONDecodeError:
                body = body
            msg += f" {json.dumps(body)}"
        super().__init__(msg)


def token_exchange(authorization_code: str, code_verifier: str) -> str:
    """Exchange an authorization code obtained from the final oauth redirect for an access token.

    Args:
        authorization_code: The authorization code obtained from the final oauth redirect.
        code_verifier: The code verifier that was originally generated before the auth request.

    Returns:
        The access token as a string.

    Raises:
        APIError: The HTTP request failed.
    """
    body = {
        "app_name": "Unknown",
        "client_domain": "DeviceLegacy",
        "client_id": "658490dfb190e494030082836775981fa23be0c2425441860352ba0f55915b43002d",
        "code_algorithm": "SHA-256",
        "code_verifier": code_verifier,
        "requested_token_type": "access_token",
        "source_token": authorization_code,
        "source_token_type": "authorization_code",
    }
    req = urllib.request.Request(
        url="https://api.amazon.com/auth/token",
        data=json.dumps(body).encode("utf-8"),
        headers={
            "Accept-Language": "en-US",
            "x-amzn-identity-auth-domain": "api.amazon.com",
            "Content-Type": "application/json",
            "User-Agent": "Mozilla/5.0",
        },
        method="POST",
    )
    try:
        with urllib.request.urlopen(req) as r:  # noqa S310
            res = json.load(r)
    except urllib.error.HTTPError as e:
        raise APIError(str(e), _text(e)) from e
    access_token: str = res["access_token"]
    return access_token


def register_device_with_token(access_token: str) -> DeviceInfo:
    """Creates a long-lived device capable of interacting with the STK API from an access_token.

    Args:
        access_token: The access token obtained from token exchange

    Returns:
        DeviceInfo instance with the newly created device.

    Raises:
        APIError: The HTTP request failed.
    """
    q = {
        "device_type": "A1K6D1WRW0MALS",
        "device_serial_number": "ZYSQ37GQ5JQDAIKDZ3WYH6I74MJCVEGG",
        "pid": "D21NN3GG",
        "auth_token": access_token,
        "auth_token_type": "AccessToken",
        "software_version": "253",
        "os_version": "MacOSX_10.14.6_x64",
        "device_model": "Maxs MacBook Pro",
    }
    body = f"""<?xml version='1.0' encoding='UTF-8'?>
<request><parameters><deviceType>{q["device_type"]}</deviceType><deviceSerialNumber>{q["device_serial_number"]}</deviceSerialNumber><pid>{q["pid"]}</pid><authToken>{q["auth_token"]}</authToken><authTokenType>{q["auth_token_type"]}</authTokenType><softwareVersion>{q["software_version"]}</softwareVersion><os_version>{q["os_version"]}</os_version><device_model>{q["device_model"]}</device_model></parameters></request>"""
    req = urllib.request.Request(
        url="https://firs-ta-g7g.amazon.com/FirsProxy/registerDeviceWithToken",
        data=body.encode(),
        headers={
            "Content-Type": "text/xml",
            "Expect": "",
            "Accept-Language": "en-US,*",
            "User-Agent": "Mozilla/5.0",
        },
        method="POST",
    )
    try:
        with urllib.request.urlopen(req) as r:  # noqa S310
            return DeviceInfo.from_xml(r.read())
    except urllib.error.HTTPError as e:
        raise APIError(str(e), _text(e)) from e


def get_list_of_owned_devices(signer: Signer) -> GetOwnedDevicesResponse:
    """Gets a list of send-to-kindle target devices.

    Args:
        signer: Signer instance to authenticate the client.

    Returns:
        GetOwnedDevicesResponse containing owned devices.

    Raises:
        APIError: The HTTP request failed.
    """
    try:
        return GetOwnedDevicesResponse.from_dict(_request("/GetListOfOwnedDevices", signer, {}))
    except urllib.error.HTTPError as e:
        raise APIError(str(e), _text(e)) from e


def get_upload_url(signer: Signer, file_size: int) -> GetUploadUrlResponse:
    """Gets a URL where the client can send the file contents via HTTP POST request.

    Args:
        signer: Signer instance to authenticate the client.
        file_size: Size of the file to be uploaded.

    Returns:
        GetUploadUrlResponse containing the upload URL and token.

    Raises:
        APIError: The HTTP request failed.
    """
    try:
        return GetUploadUrlResponse.from_dict(
            _request("/GetUploadUrl", signer, {"fileSize": file_size})
        )
    except urllib.error.HTTPError as e:
        raise APIError(str(e), _text(e)) from e


def upload_file(url: str, file_size: int, fp: IO[Any]) -> None:
    """Perform a streaming upload of a file to the supplied URL via HTTP PUT request.

    Args:
        url: Where to upload the file
        file_size: Size of the file to be uploaded.
        fp: Readable binary file-like object to upload.

    Raises:
        ValueError: The supplied URL is invalid.
        APIError: The HTTP request failed.
    """
    u = urllib.parse.urlparse(url)
    if u.hostname is None:
        raise ValueError("Invalid URL")
    conn = http.client.HTTPSConnection(u.hostname)  # noqa: S309
    try:
        headers = {
            "Accept-Encoding": "gzip, deflate",
            "Accept-Language": "en-US,*",
            "Content-Length": str(file_size),
            "User-Agent": "Mozilla/5.0",
        }
        conn.request(
            "PUT",
            url,
            body=fp,
            headers=headers,
        )
        res = conn.getresponse()
        text = res.read()
        if res.status != 200:
            msg = f"HTTP Status Error {res.status} {res.reason}"
            raise APIError(msg, text)
    finally:
        conn.close()


def send_to_kindle(
    signer: Signer,
    stk_token: str,
    target_device_serial_numbers: List[str],
    *,
    author: str,
    title: str,
    format: str,
) -> SendToKindleResponse:
    """Send an uploaded file to the specified kindle devices.

    Args:
        signer: Signer instance to authenticate the client.
        stk_token: The token associated with the upload url.
        target_device_serial_numbers: The devices to receive the file.
        author: The author of the document.
        title: The title of the document.
        format: The format of the document.

    Returns:
        SendToKindleResponse containing metadata about the sent file.

    Raises:
        APIError: The HTTP request failed.
    """
    body = {
        "DocumentMetadata": {
            "author": author,
            "crc32": 0,
            "inputFormat": format,
            "title": title,
        },
        "archive": True,
        "deliveryMechanism": "WIFI",
        "outputFormat": "MOBI",
        "stkToken": stk_token,
        "targetDevices": target_device_serial_numbers,
    }
    try:
        return SendToKindleResponse.from_dict(_request("/SendToKindle", signer, body))
    except urllib.error.HTTPError as e:
        raise APIError(str(e), _text(e)) from e


def logout(signer: Signer) -> None:
    """Logs out a send-to-kindle client.

    Args:
        signer: Signer instance to authenticate the client.

    Raises:
        APIError: The HTTP request failed.
    """
    path = "/FirsProxy/disownFiona?contentDeleted=false"
    req = urllib.request.Request(
        url="https://firs-ta-g7g.amazon.com" + path,
        headers={
            "Content-Type": "text/xml",
            "X-ADP-Request-Digest": signer.digest_header_for_request("GET", path, ""),
            "X-ADP-Authentication-Token": signer.adp_token,
            "Accept-Language": "en-US,*",
            "User-Agent": "Mozilla/5.0",
        },
        method="GET",
    )
    try:
        with urllib.request.urlopen(req) as r:  # noqa S310
            r.read()  # Read and discard
    except urllib.error.HTTPError as e:
        raise APIError(str(e), _text(e)) from e


def _request(path: str, signer: Signer, body: Mapping[str, Any]) -> Mapping[str, Any]:
    data = json.dumps(
        {
            "ClientInfo": DEFAULT_CLIENT_INFO,
            **body,
        },
        indent=4,
    )
    req = urllib.request.Request(
        url="https://stkservice.amazon.com" + path,
        data=data.encode("utf-8"),
        headers={
            "Accept": "application/json",
            "Accept-Encoding": "gzip, deflate",
            "Content-Type": "application/json",
            "X-ADP-Request-Digest": signer.digest_header_for_request("POST", path, data),
            "X-ADP-Authentication-Token": signer.adp_token,
            "Accept-Language": "en-US,*",
            "User-Agent": "Mozilla/5.0",
        },
        method="POST",
    )
    with urllib.request.urlopen(req) as r:  # noqa S310
        val: Mapping[str, Any] = json.load(r)
        return val


def _text(e: urllib.error.HTTPError) -> Optional[bytes]:
    try:
        return e.read()
    except AttributeError:
        return None



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['stkclient']

package_data = \
{'': ['*']}

install_requires = \
['defusedxml>=0.7.1,<0.8.0', 'rsa>=4.8,<5.0']

entry_points = \
{'console_scripts': ['stkclient = stkclient.__main__:main']}

setup_kwargs = {
    'name': 'stkclient',
    'version': '0.1.0',
    'description': 'Send To Kindle',
    'long_description': 'Send To Kindle\n==============\n\n|PyPI| |Status| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/stkclient.svg\n   :target: https://pypi.org/project/stkclient/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/stkclient.svg\n   :target: https://pypi.org/project/stkclient/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/stkclient\n   :target: https://pypi.org/project/stkclient\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/stkclient\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/stkclient/latest.svg?label=Read%20the%20Docs\n   :target: https://stkclient.readthedocs.io/\n   :alt: Read the documentation at https://stkclient.readthedocs.io/\n.. |Tests| image:: https://github.com/maxdjohnson/stkclient/workflows/Tests/badge.svg\n   :target: https://github.com/maxdjohnson/stkclient/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/maxdjohnson/stkclient/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/maxdjohnson/stkclient\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n\n``stkclient`` implements a client for amazon\'s "Send to Kindle" service. It allows python programs to\nsend files to a kindle device without the 10mb limit that applies to email files.\n\nFeatures\n--------\n\n* OAuth-based authorization\n* Send large (>10MB) files to Kindle devices\n\n\nRequirements\n------------\n\n* TODO\n\n\nInstallation\n------------\n\nYou can install *Send To Kindle* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install stkclient\n\n\nCreating a Client\n-----------------\n\nTo create a client, you must authenticate the user. Currently the only supported authentication mechanism is OAuth2:\n\n.. code:: python\n\n   import stkclient\n\n   a = stkclient.OAuth2()\n   signin_url = a.get_signin_url()\n   # Open `signin_url` in a browser, sign in and authorize the application, pass\n   # the final redirect_url below\n   client = a.create_client(redirect_url)\n\nOnce a client is created, it can be serialized and deserialized using ``Client.load`` / ``Client.loads`` and ``client.dump`` / ``client.dumps``\n\n.. code:: python\n\n   with open(\'client.json\', \'w\') as f:\n       client.dump(f)\n   with open(\'client.json\', \'r\') as f:\n       client = stkclient.Client.load(f)\n\n\nSending a File\n--------------\n\nOnce you have a Client object, you can list devices and send files to specified devices.\n\n.. code:: python\n\n   devices = client.get_owned_devices()\n   destinations = [d.device_serial_number for d in devices.owned_devices]\n   client.send_file(filepath, destinations, author=author, title=title)\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*Send To Kindle* is free and open source software.\n\n\nCredits\n-------\n\nProject structure from `@cjolowicz`_\'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _pip: https://pip.pypa.io/\n',
    'author': 'Max Johnson',
    'author_email': 'maxdjohnson@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/maxdjohnson/stkclient',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Command-line interface."""
import argparse
import os
import sys
from pathlib import Path
from typing import List, Optional

import stkclient

# Try to import the readline module for improved input() behavior. Without this, pasting a line
# longer than 1024 chars causes the process to hang on my machine.
try:
    import readline  # noqa
except ModuleNotFoundError:
    pass  # not available on windows

DEFAULT_CLIENT_PATH = os.path.join("$XDG_DATA_HOME", "pystkclient", "client.json")


def arg_parser() -> argparse.ArgumentParser:
    """Constructs an ArgumentParser for the stkclient script."""
    # create the top-level parser
    parser = argparse.ArgumentParser(
        prog="stkclient", description="Command-line interface for Amazon's Send-to-Kindle service"
    )
    subparsers = parser.add_subparsers()

    # create the parser for the "login" command
    parser_login = subparsers.add_parser("login", help=login.__doc__)
    parser_login.add_argument(
        "--client",
        type=str,
        default=DEFAULT_CLIENT_PATH,
        help="path to write the client details",
    )
    parser_login.add_argument(
        "-f", "--force", action="store_true", help="overwrite the existing client, if there is one"
    )
    parser_login.set_defaults(func=login)

    # create the parser for the "devices" command
    parser_devices = subparsers.add_parser("devices", help=devices.__doc__)
    parser_devices.add_argument(
        "--client",
        type=str,
        default=DEFAULT_CLIENT_PATH,
        help="path to the client details",
    )
    parser_devices.set_defaults(func=devices)

    # create the parser for the "send" command
    parser_send = subparsers.add_parser("send", help=send.__doc__)
    parser_send.add_argument(
        "--client",
        type=str,
        default=DEFAULT_CLIENT_PATH,
        help="path to the client details",
    )
    parser_send.add_argument(
        "--title", type=str, required=True, help="title of the work (required)"
    )
    parser_send.add_argument(
        "--author", type=str, required=True, help="author of the work (required)"
    )
    parser_send.add_argument(
        "--format", type=str, required=True, help='file format, for example "mobi" (required)'
    )
    parser_send.add_argument("file", type=Path, help="file to send")
    parser_send.add_argument(
        "target",
        type=str,
        nargs="+",
        help='device serial numbers to send the file to, or "all" to send to all devices',
    )
    parser_send.set_defaults(func=send)

    # create the parser for the "logout" command
    parser_logout = subparsers.add_parser("logout", help=logout.__doc__)
    parser_logout.add_argument(
        "--client",
        type=str,
        default=DEFAULT_CLIENT_PATH,
        help="path to the client details - will be deleted",
    )
    parser_logout.set_defaults(func=logout)

    # TODO verbosity
    return parser


def main(args: Optional[List[str]] = None) -> None:
    """Send To Kindle."""
    parser = arg_parser()
    parsed = parser.parse_args(args)
    if not hasattr(parsed, "func"):
        parser.print_usage()
        exit(1)
    parsed.func(parsed)


def login(args: argparse.Namespace) -> None:
    """Create a client by authenticating with OAuth2."""
    client_path = _get_client_path(args)
    if client_path.exists() and not args.force:
        print(f"{client_path} already exists", file=sys.stderr)
        exit(1)
    if args.client == DEFAULT_CLIENT_PATH:
        client_path.parent.mkdir(parents=True, exist_ok=True)
    elif not client_path.parent.is_dir():
        print(f"{client_path.parent} is not a directory", file=sys.stderr)
        exit(1)
    auth = stkclient.OAuth2()
    signin_url = auth.get_signin_url()
    print(signin_url)
    redirect_url = input("Enter redirect url: ")
    client = auth.create_client(redirect_url)
    with open(client_path, "w") as f:
        client.dump(f)


def devices(args: argparse.Namespace) -> None:
    """List available kindle reader devices."""
    client_path = _get_client_path(args)
    if not client_path.exists():
        print(f"{client_path} does not exist", file=sys.stderr)
        exit(1)
    with open(client_path) as f:
        client = stkclient.Client.load(f)
    devices = client.get_owned_devices()
    for device in devices:
        print(f"{device.device_serial_number}: {device.device_name}")


def send(args: argparse.Namespace) -> None:
    """Send a file to one or more devices."""
    client_path = _get_client_path(args)
    if not client_path.exists():
        print(f"{client_path} does not exist", file=sys.stderr)
        exit(1)
    with open(client_path) as f:
        client = stkclient.Client.load(f)
    target: List[str] = args.target
    if any(t == "all" for t in args.target):
        target = [d.device_serial_number for d in client.get_owned_devices()]
    client.send_file(args.file, target, author=args.author, title=args.title, format=args.format)


def logout(args: argparse.Namespace) -> None:
    """Deauthorize and delete a client."""
    client_path = _get_client_path(args)
    if not client_path.exists():
        print(f"{client_path} does not exist", file=sys.stderr)
        exit(1)
    with open(client_path) as f:
        c = stkclient.Client.load(f)
    c.logout()
    client_path.unlink()


def _get_client_path(args: argparse.Namespace) -> Path:
    client_path: str = args.client
    data_home = os.environ.get("XDG_DATA_HOME", os.path.join("~", ".local", "share"))
    return Path(client_path.replace("$XDG_DATA_HOME", data_home)).expanduser()


if __name__ == "__main__":
    main()  # pragma: no cover
